from typing import Annotated

from pydantic import BaseModel, Field, SecretStr, StrictStr


class S3Coordinates(BaseModel):
    endpoint: Annotated[
        StrictStr,
        Field(
            description=("S3 service endpoint, without protocol. Required."),
            examples=[
                "s3.eu-de.cloud-object-storage.appdomain.cloud",
                "s3.us-east-2.amazonaws.com ",
            ],
        ),
    ]

    verify_ssl: Annotated[
        bool,
        Field(
            description=(
                "If enabled, SSL will be used to connect to s3. "
                "Boolean. Optional, defaults to true"
            ),
        ),
    ] = True

    access_key: Annotated[
        SecretStr,
        Field(
            description=("S3 access key. Required."),
        ),
    ]

    secret_key: Annotated[
        SecretStr,
        Field(
            description=("S3 secret key. Required."),
        ),
    ]

    bucket: Annotated[
        str,
        Field(
            description=("S3 bucket name. Required."),
        ),
    ]

    key_prefix: Annotated[
        str,
        Field(
            description=(
                "Prefix for the object keys on s3. Optional, defaults to empty."
            ),
        ),
    ] = ""



import datetime
import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Optional, Union

from docling.datamodel.document import ConversionResult

from docling_jobkit.datamodel.callback import ProgressCallbackRequest
from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
from docling_jobkit.datamodel.task import Task, TaskSource

if TYPE_CHECKING:
    from docling_jobkit.orchestrators.base_notifier import BaseNotifier

_log = logging.getLogger(__name__)


class OrchestratorError(Exception):
    pass


class TaskNotFoundError(OrchestratorError):
    pass


class ProgressInvalid(OrchestratorError):
    pass


class BaseOrchestrator(ABC):
    def __init__(self):
        self.tasks: dict[str, Task] = {}
        self.notifier: Optional["BaseNotifier"] = None

    def bind_notifier(self, notifier: "BaseNotifier"):
        self.notifier = notifier

    @abstractmethod
    async def enqueue(
        self, sources: list[TaskSource], options: ConvertDocumentsOptions
    ) -> Task:
        pass

    @abstractmethod
    async def queue_size(self) -> int:
        pass

    @abstractmethod
    async def get_queue_position(self, task_id: str) -> Optional[int]:
        pass

    @abstractmethod
    async def process_queue(self):
        pass

    @abstractmethod
    async def warm_up_caches(self):
        pass

    @abstractmethod
    async def clear_converters(self):
        pass

    async def init_task_tracking(self, task: Task):
        task_id = task.task_id
        self.tasks[task.task_id] = task
        if self.notifier:
            await self.notifier.add_task(task_id)

    async def get_raw_task(self, task_id: str) -> Task:
        if task_id not in self.tasks:
            raise TaskNotFoundError()
        return self.tasks[task_id]

    async def task_status(self, task_id: str, wait: float = 0.0) -> Task:
        return await self.get_raw_task(task_id=task_id)

    async def task_results(
        # self, task_id: str, background_tasks: BackgroundTasks
        self,
        task_id: str,
    ) -> Union[list[ConversionResult], None]:
        try:
            task = await self.get_raw_task(task_id=task_id)
            return task.results
        except TaskNotFoundError:
            return None

    async def delete_task(self, task_id: str):
        _log.info(f"Deleting {task_id=}")
        if self.notifier:
            await self.notifier.remove_task(task_id=task_id)
        if task_id in self.tasks:
            del self.tasks[task_id]

    async def clear_results(self, older_than: float = 0.0):
        cutoff_time = datetime.datetime.now(datetime.timezone.utc) - datetime.timedelta(
            seconds=older_than
        )

        tasks_to_delete = [
            task_id
            for task_id, task in self.tasks.items()
            if task.finished_at is not None and task.finished_at < cutoff_time
        ]
        for task_id in tasks_to_delete:
            await self.delete_task(task_id=task_id)

    async def receive_task_progress(self, request: ProgressCallbackRequest):
        raise NotImplementedError()



import enum
import hashlib
import json
import logging
import re
import sys
from collections.abc import Iterable, Iterator
from functools import lru_cache
from pathlib import Path
from typing import Any, Callable, Optional, Union

from pydantic import BaseModel

from docling.backend.docling_parse_backend import DoclingParseDocumentBackend
from docling.backend.docling_parse_v2_backend import DoclingParseV2DocumentBackend
from docling.backend.docling_parse_v4_backend import DoclingParseV4DocumentBackend
from docling.backend.pdf_backend import PdfDocumentBackend
from docling.backend.pypdfium2_backend import PyPdfiumDocumentBackend
from docling.datamodel import vlm_model_specs
from docling.datamodel.base_models import DocumentStream, InputFormat
from docling.datamodel.document import ConversionResult
from docling.datamodel.pipeline_options import (
    OcrOptions,
    PdfBackend,
    PdfPipelineOptions,
    PictureDescriptionApiOptions,
    PictureDescriptionVlmOptions,
    ProcessingPipeline,
    TableFormerMode,
    VlmPipelineOptions,
)
from docling.document_converter import DocumentConverter, FormatOption, PdfFormatOption
from docling.models.factories import get_ocr_factory
from docling.pipeline.vlm_pipeline import VlmPipeline
from docling_core.types.doc import ImageRefMode

from docling_jobkit.datamodel.convert import ConvertDocumentsOptions

_log = logging.getLogger(__name__)


class DoclingConverterManagerConfig(BaseModel):
    artifacts_path: Optional[Path] = None
    options_cache_size: int = 2
    enable_remote_services: bool = False
    allow_external_plugins: bool = False

    max_num_pages: int = sys.maxsize
    max_file_size: int = sys.maxsize


# Custom serializer for PdfFormatOption
# (model_dump_json does not work with some classes)
def _hash_pdf_format_option(pdf_format_option: PdfFormatOption) -> bytes:
    data = pdf_format_option.model_dump(serialize_as_any=True)

    # pipeline_options are not fully serialized by model_dump, dedicated pass
    if pdf_format_option.pipeline_options:
        data["pipeline_options"] = pdf_format_option.pipeline_options.model_dump(
            serialize_as_any=True, mode="json"
        )

    # Replace `pipeline_cls` with a string representation
    data["pipeline_cls"] = repr(data["pipeline_cls"])

    # Replace `backend` with a string representation
    data["backend"] = repr(data["backend"])

    # Serialize the dictionary to JSON with sorted keys to have consistent hashes
    serialized_data = json.dumps(data, sort_keys=True)
    options_hash = hashlib.sha1(
        serialized_data.encode(), usedforsecurity=False
    ).digest()
    return options_hash


def _to_list_of_strings(input_value: Union[str, list[str]]) -> list[str]:
    def split_and_strip(value: str) -> list[str]:
        if re.search(r"[;,]", value):
            return [item.strip() for item in re.split(r"[;,]", value)]
        else:
            return [value.strip()]

    if isinstance(input_value, str):
        return split_and_strip(input_value)
    elif isinstance(input_value, list):
        result = []
        for item in input_value:
            result.extend(split_and_strip(str(item)))
        return result
    else:
        raise ValueError("Invalid input: must be a string or a list of strings.")


class DoclingConverterManager:
    def __init__(self, config: DoclingConverterManagerConfig):
        self.config = config

        self.ocr_factory = get_ocr_factory(
            allow_external_plugins=self.config.allow_external_plugins
        )
        self._options_map: dict[bytes, PdfFormatOption] = {}
        self._get_converter_from_hash = self._create_converter_cache_from_hash(
            cache_size=self.config.options_cache_size
        )

    def _create_converter_cache_from_hash(
        self, cache_size: int
    ) -> Callable[[bytes], DocumentConverter]:
        @lru_cache(maxsize=cache_size)
        def _get_converter_from_hash(options_hash: bytes) -> DocumentConverter:
            pdf_format_option = self._options_map[options_hash]
            format_options: dict[InputFormat, FormatOption] = {
                InputFormat.PDF: pdf_format_option,
                InputFormat.IMAGE: pdf_format_option,
            }

            return DocumentConverter(format_options=format_options)

        return _get_converter_from_hash

    def clear_cache(self):
        self._get_converter_from_hash.cache_clear()

    def get_converter(self, pdf_format_option: PdfFormatOption) -> DocumentConverter:
        options_hash = _hash_pdf_format_option(pdf_format_option)
        self._options_map[options_hash] = pdf_format_option
        return self._get_converter_from_hash(options_hash)

    def _parse_standard_pdf_opts(
        self, request: ConvertDocumentsOptions, artifacts_path: Optional[Path]
    ) -> PdfPipelineOptions:
        try:
            kind = (
                request.ocr_engine.value
                if isinstance(request.ocr_engine, enum.Enum)
                else str(request.ocr_engine)
            )
            ocr_options: OcrOptions = self.ocr_factory.create_options(  # type: ignore
                kind=kind,
                force_full_page_ocr=request.force_ocr,
            )
        except ImportError as err:
            raise ImportError(
                "The requested OCR engine"
                f" (ocr_engine={request.ocr_engine})"
                " is not available on this system. Please choose another OCR engine "
                "or contact your system administrator.\n"
                f"{err}"
            )

        if request.ocr_lang is not None:
            ocr_options.lang = request.ocr_lang

        pipeline_options = PdfPipelineOptions(
            artifacts_path=artifacts_path,
            enable_remote_services=self.config.enable_remote_services,
            document_timeout=request.document_timeout,
            do_ocr=request.do_ocr,
            ocr_options=ocr_options,
            do_table_structure=request.do_table_structure,
            do_code_enrichment=request.do_code_enrichment,
            do_formula_enrichment=request.do_formula_enrichment,
            do_picture_classification=request.do_picture_classification,
            do_picture_description=request.do_picture_description,
        )
        pipeline_options.table_structure_options.mode = TableFormerMode(
            request.table_mode
        )

        if request.image_export_mode != ImageRefMode.PLACEHOLDER:
            pipeline_options.generate_page_images = True
            if request.image_export_mode == ImageRefMode.REFERENCED:
                pipeline_options.generate_picture_images = True
            if request.images_scale:
                pipeline_options.images_scale = request.images_scale

        if request.picture_description_local is not None:
            pipeline_options.picture_description_options = (
                PictureDescriptionVlmOptions.model_validate(
                    request.picture_description_local.model_dump()
                )
            )

        if request.picture_description_api is not None:
            pipeline_options.picture_description_options = (
                PictureDescriptionApiOptions.model_validate(
                    request.picture_description_api.model_dump()
                )
            )
        pipeline_options.picture_description_options.picture_area_threshold = (
            request.picture_description_area_threshold
        )

        return pipeline_options

    def _parse_backend(
        self, request: ConvertDocumentsOptions
    ) -> type[PdfDocumentBackend]:
        if request.pdf_backend == PdfBackend.DLPARSE_V1:
            backend: type[PdfDocumentBackend] = DoclingParseDocumentBackend
        elif request.pdf_backend == PdfBackend.DLPARSE_V2:
            backend = DoclingParseV2DocumentBackend
        elif request.pdf_backend == PdfBackend.DLPARSE_V4:
            backend = DoclingParseV4DocumentBackend
        elif request.pdf_backend == PdfBackend.PYPDFIUM2:
            backend = PyPdfiumDocumentBackend
        else:
            raise RuntimeError(f"Unexpected PDF backend type {request.pdf_backend}")

        return backend

    def _parse_vlm_pdf_opts(
        self, request: ConvertDocumentsOptions, artifacts_path: Optional[Path]
    ) -> VlmPipelineOptions:
        pipeline_options = VlmPipelineOptions(
            artifacts_path=artifacts_path,
            document_timeout=request.document_timeout,
        )
        pipeline_options.vlm_options = vlm_model_specs.SMOLDOCLING_TRANSFORMERS
        if sys.platform == "darwin":
            try:
                import mlx_vlm  # noqa: F401

                pipeline_options.vlm_options = vlm_model_specs.SMOLDOCLING_MLX
            except ImportError:
                _log.warning(
                    "To run SmolDocling faster, please install mlx-vlm:\n"
                    "pip install mlx-vlm"
                )
        return pipeline_options

    # Computes the PDF pipeline options and returns the PdfFormatOption and its hash
    def get_pdf_pipeline_opts(
        self,
        request: ConvertDocumentsOptions,
    ) -> PdfFormatOption:
        artifacts_path: Optional[Path] = None
        if self.config.artifacts_path is not None:
            expanded_path = self.config.artifacts_path.expanduser()
            if str(expanded_path.absolute()) == "":
                _log.info(
                    "artifacts_path is an empty path, model weights will be downloaded "
                    "at runtime."
                )
                artifacts_path = None
            elif expanded_path.is_dir():
                _log.info(
                    "artifacts_path is set to a valid directory. "
                    "No model weights will be downloaded at runtime."
                )
                artifacts_path = expanded_path
            else:
                _log.warning(
                    "artifacts_path is set to an invalid directory. "
                    "The system will download the model weights at runtime."
                )
                artifacts_path = None
        else:
            _log.info(
                "artifacts_path is unset. "
                "The system will download the model weights at runtime."
            )

        pipeline_options: Union[PdfPipelineOptions, VlmPipelineOptions]
        if request.pipeline == ProcessingPipeline.STANDARD:
            pipeline_options = self._parse_standard_pdf_opts(request, artifacts_path)
            backend = self._parse_backend(request)
            pdf_format_option = PdfFormatOption(
                pipeline_options=pipeline_options,
                backend=backend,
            )

        elif request.pipeline == ProcessingPipeline.VLM:
            pipeline_options = self._parse_vlm_pdf_opts(request, artifacts_path)
            pdf_format_option = PdfFormatOption(
                pipeline_cls=VlmPipeline, pipeline_options=pipeline_options
            )
        else:
            raise NotImplementedError(
                f"The pipeline {request.pipeline} is not implemented."
            )

        return pdf_format_option

    def convert_documents(
        self,
        sources: Iterable[Union[Path, str, DocumentStream]],
        options: ConvertDocumentsOptions,
        headers: Optional[dict[str, Any]] = None,
    ) -> Iterable[ConversionResult]:
        pdf_format_option = self.get_pdf_pipeline_opts(options)
        converter = self.get_converter(pdf_format_option)
        results: Iterator[ConversionResult] = converter.convert_all(
            sources,
            headers=headers,
            page_range=options.page_range,
            max_file_size=self.config.max_file_size,
            max_num_pages=self.config.max_num_pages,
        )

        return results



import argparse
import json
import os
from typing import Optional
from urllib.parse import urlparse, urlunsplit

import ray
from boto3.resources.base import ServiceResource
from boto3.session import Session
from botocore.client import BaseClient
from botocore.config import Config
from botocore.exceptions import ClientError
from botocore.paginate import Paginator
from pydantic import BaseModel
from ray._raylet import ObjectRefGenerator  # type: ignore

from docling.backend.docling_parse_backend import DoclingParseDocumentBackend
from docling.datamodel.base_models import ConversionStatus, InputFormat
from docling.datamodel.document import ConversionResult
from docling.datamodel.pipeline_options import PdfPipelineOptions, TableFormerMode
from docling.document_converter import DocumentConverter, PdfFormatOption

# Load credentials
s3_source_access_key = os.environ["S3_SOURCE_ACCESS_KEY"]
s3_source_secret_key = os.environ["S3_SOURCE_SECRET_KEY"]
s3_source_endpoint = os.environ["S3_SOURCE_ENDPOINTS"]
s3_source_bucket = os.environ["S3_SOURCE_BUCKET"]
s3_source_prefix = os.environ["S3_SOURCE_PREFIX"]
s3_source_ssl = os.environ.get("S3_SOURCE_SSL", True)
s3_target_access_key = os.environ["S3_TARGET_ACCESS_KEY"]
s3_target_secret_key = os.environ["S3_TARGET_SECRET_KEY"]
s3_target_endpoint = os.environ["S3_TARGET_ENDPOINTS"]
s3_target_bucket = os.environ["S3_TARGET_BUCKET"]
s3_target_prefix = os.environ["S3_TARGET_PREFIX"]
s3_target_ssl = os.environ.get("S3_TARGET_SSL", True)
batch_size = int(os.environ["BATCH_SIZE"])
max_concurrency = int(os.environ["OMP_NUM_THREADS"])

# Load conversion settings
do_ocr = bool(os.environ.get("SETTINGS_DO_OCR", True))
do_table_structure = bool(os.environ.get("SETTINGS_DO_TABLE_STRUCTURE", True))
table_structure_mode = os.environ.get("SETTINGS_TABLE_STRUCTURE_MODE", "fast")
generate_page_images = bool(os.environ.get("SETTINGS_GENERATE_PAGE_IMAGES", True))


class S3Coordinates(BaseModel):
    endpoint: str
    verify_ssl: bool
    access_key: str
    secret_key: str
    bucket: str
    key_prefix: str


def get_s3_connection(coords: S3Coordinates):
    session = Session()

    config = Config(
        connect_timeout=30, retries={"max_attempts": 1}, signature_version="s3v4"
    )
    scheme = "https" if coords.verify_ssl else "http"
    path = "/"
    endpoint = urlunsplit((scheme, coords.endpoint, path, "", ""))

    client: BaseClient = session.client(
        "s3",
        endpoint_url=endpoint,
        verify=coords.verify_ssl,
        aws_access_key_id=coords.access_key,
        aws_secret_access_key=coords.secret_key,
        config=config,
    )

    resource: ServiceResource = session.resource(
        "s3",
        endpoint_url=endpoint,
        verify=coords.verify_ssl,
        aws_access_key_id=coords.access_key,
        aws_secret_access_key=coords.secret_key,
        config=config,
    )

    return client, resource


def count_s3_objects(paginator: Paginator, bucket_name: str, prefix: str):
    response_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    count_obj = 0
    for page in response_iterator:
        if page.get("Contents"):
            count_obj += sum(1 for _ in page["Contents"])

    return count_obj


def get_keys_s3_objects_as_set(
    s3_resource: ServiceResource, bucket_name: str, prefix: str
):
    bucket = s3_resource.Bucket(bucket_name)
    folder_objects = list(bucket.objects.filter(Prefix=prefix))
    files_on_s3 = set()
    for file in folder_objects:
        files_on_s3.add(file.key)
    return files_on_s3


def strip_prefix_postfix(source_set, prefix="", extension=""):
    output = set()
    for key in source_set:
        output.add(key.replace(extension, "").replace(prefix, ""))
    return output


def generate_presigns_url(s3_client: BaseClient, source_keys: list):
    presigned_urls = []
    counter = 0
    sub_array = []
    array_lenght = len(source_keys)
    for idx, key in enumerate(source_keys):
        try:
            url = s3_client.generate_presigned_url(
                ClientMethod="get_object",
                Params={"Bucket": s3_source_bucket, "Key": key},
                ExpiresIn=3600,
            )
        except ClientError as e:
            print(e)
        sub_array.append(url)
        counter += 1
        if counter == batch_size or (idx + 1) == array_lenght:
            presigned_urls.append(sub_array)
            sub_array = []
            counter = 0

    return presigned_urls


def get_source_files(s3_source_client, s3_source_resource):
    source_paginator = s3_source_client.get_paginator("list_objects_v2")

    # Check that source is not empty
    source_count = count_s3_objects(
        source_paginator, s3_source_bucket, s3_source_prefix + "/"
    )
    if source_count == 0:
        print("s3 source is empty")
        ray.shutdown()
    return get_keys_s3_objects_as_set(
        s3_source_resource, s3_source_bucket, s3_source_prefix
    )


def check_target_has_source_converted(coords, source_objects_list):
    s3_target_client, s3_target_resource = get_s3_connection(coords)
    target_paginator = s3_target_client.get_paginator("list_objects_v2")

    converted_prefix = s3_target_prefix + "/json/"
    target_count = count_s3_objects(
        target_paginator, s3_target_bucket, converted_prefix
    )
    print("Target contains json objects: ", target_count)
    if target_count != 0:
        print("Target contains objects, checking content...")

        # Collect target keys for iterative conversion
        existing_target_objects = get_keys_s3_objects_as_set(
            s3_target_resource, s3_target_bucket, converted_prefix
        )

        # Filter-out objects that are already processed
        target_short_key_list = strip_prefix_postfix(
            existing_target_objects, prefix=converted_prefix, extension=".json"
        )
        filtered_source_keys = []
        print("List of source keys:")
        for key in source_objects_list:
            print(key)
            clean_key = key.replace(".pdf", "").replace(s3_source_prefix + "/", "")
            if clean_key not in target_short_key_list:
                filtered_source_keys.append(key)

        print("Total keys: ", len(source_objects_list))
        print("Filtered keys to process: ", len(filtered_source_keys))
    else:
        filtered_source_keys = source_objects_list

    return filtered_source_keys


def put_object(
    client,
    bucket: str,
    object_key: str,
    file: str,
    content_type: Optional[str] = None,
) -> bool:
    """Upload a file to an S3 bucket

    :param file: File to upload
    :param bucket: Bucket to upload to
    :param object_key: S3 key to upload to
    :return: True if file was uploaded, else False
    """

    kwargs = {}

    if content_type is not None:
        kwargs["ContentType"] = content_type

    try:
        client.put_object(Body=file, Bucket=bucket, Key=object_key, **kwargs)
    except ClientError as e:
        print(e)
        return False
    return True


@ray.remote(max_concurrency=max_concurrency)  # type: ignore
class DoclingConvert:
    def __init__(self, s3_coords: S3Coordinates, presigned_urls: list):
        self.coords = s3_coords
        self.s3_client, _ = get_s3_connection(s3_coords)
        self.presigned_urls = presigned_urls

        pipeline_options = PdfPipelineOptions()
        pipeline_options.do_ocr = do_ocr
        pipeline_options.do_table_structure = do_table_structure
        pipeline_options.table_structure_options.mode = TableFormerMode(
            table_structure_mode
        )
        pipeline_options.generate_page_images = generate_page_images

        self.converter = DocumentConverter(
            format_options={
                InputFormat.PDF: PdfFormatOption(
                    pipeline_options=pipeline_options,
                    backend=DoclingParseDocumentBackend,
                )
            }
        )
        self.allowed_formats = [ext.value for ext in InputFormat]

    def convert_document(self, index, db_ref):
        for url in db_ref[index]:
            parsed = urlparse(url)
            root, ext = os.path.splitext(parsed.path)
            if ext[1:] not in self.allowed_formats:
                continue
            conv_res: ConversionResult = self.converter.convert(url)
            if conv_res.status == ConversionStatus.SUCCESS:
                doc_filename = conv_res.input.file.stem
                print(f"Converted {doc_filename} now saving results")
                # Export Docling document format to JSON:
                target_key = f"{s3_target_prefix}/json/{doc_filename}.json"
                data = json.dumps(conv_res.document.export_to_dict())
                self.upload_to_s3(
                    file=data,
                    target_key=target_key,
                    content_type="application/json",
                )
                # Export Docling document format to doctags:
                target_key = f"{s3_target_prefix}/doctags/{doc_filename}.doctags.txt"
                data = conv_res.document.export_to_document_tokens()
                self.upload_to_s3(
                    file=data,
                    target_key=target_key,
                    content_type="text/plain",
                )
                # Export Docling document format to markdown:
                target_key = f"{s3_target_prefix}/md/{doc_filename}.md"
                data = conv_res.document.export_to_markdown()
                self.upload_to_s3(
                    file=data,
                    target_key=target_key,
                    content_type="text/markdown",
                )
                # Export Docling document format to text:
                target_key = f"{s3_target_prefix}/txt/{doc_filename}.txt"
                data = conv_res.document.export_to_markdown(strict_text=True)
                self.upload_to_s3(
                    file=data,
                    target_key=target_key,
                    content_type="text/plain",
                )
                yield f"{doc_filename} - SUCCESS"

            elif conv_res.status == ConversionStatus.PARTIAL_SUCCESS:
                yield f"{conv_res.input.file} - PARTIAL_SUCCESS"
            else:
                yield f"{conv_res.input.file} - FAILURE"

    def upload_to_s3(self, file, target_key, content_type):
        return put_object(
            client=self.s3_client,
            bucket=self.coords.bucket,
            object_key=target_key,
            file=file,
            content_type=content_type,
        )


# This is executed on the ray-head
def main(args):
    # Init ray
    ray.init(local_mode=False)

    # Check inputs
    if (
        (not s3_source_access_key)
        or (not s3_source_secret_key)
        or (not s3_target_access_key)
        or (not s3_target_secret_key)
    ):
        print("s3 source or target keys are missing")
        ray.shutdown()
    if (not s3_source_endpoint) or (not s3_target_endpoint):
        print("s3 source or target endpoint is missing")
        ray.shutdown()
    if (not s3_source_bucket) or (not s3_target_bucket):
        print("s3 source or target bucket is missing")
        ray.shutdown()
    if (
        (s3_source_endpoint == s3_target_endpoint)
        and (s3_source_bucket == s3_target_bucket)
        and (s3_source_prefix == s3_target_prefix)
    ):
        print("s3 source and target are the same")
        ray.shutdown()
    if batch_size == 0:
        print("batch_size have to be higher than zero")
        ray.shutdown()

    # get source keys
    s3_coords_source = S3Coordinates(
        endpoint=s3_source_endpoint,
        verify_ssl=s3_source_ssl,
        access_key=s3_source_access_key,
        secret_key=s3_source_secret_key,
        bucket=s3_source_bucket,
        key_prefix=s3_source_prefix,
    )
    s3_source_client, s3_source_resource = get_s3_connection(s3_coords_source)
    source_objects_list = get_source_files(s3_source_client, s3_source_resource)

    # filter source keys
    s3_coords_target = S3Coordinates(
        endpoint=s3_target_endpoint,
        verify_ssl=s3_target_ssl,
        access_key=s3_target_access_key,
        secret_key=s3_target_secret_key,
        bucket=s3_target_bucket,
        key_prefix=s3_target_prefix,
    )

    filtered_source_keys = check_target_has_source_converted(
        s3_coords_target, source_objects_list
    )

    # Generate pre-signed urls
    presigned_urls = generate_presigns_url(s3_source_client, filtered_source_keys)

    # Init ray actor
    c = DoclingConvert.remote(s3_coords_target, presigned_urls)
    # Send payload to ray
    db_object_ref = ray.put(presigned_urls)
    # Launch tasks
    object_references = [
        c.convert_document.remote(index, db_object_ref)
        for index in range(len(presigned_urls))
    ]

    ready, unready = [], object_references
    result = []
    while unready:
        ready, unready = ray.wait(unready)
        for r in ready:
            if isinstance(r, ObjectRefGenerator):
                try:
                    ref = next(r)
                    result.append(ray.get(ref))
                except StopIteration:
                    pass
                else:
                    print("Unready")
                    unready.append(r)
            else:
                result.append(ray.get(r))

    print(result)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Basic docling ray app")

    args = parser.parse_args()
    main(args)



import json
import logging
import os
import tempfile
from datetime import datetime
from io import BytesIO
from pathlib import Path
from typing import Any
from urllib.parse import urlparse, urlunsplit

import httpx
import pandas as pd
from boto3.resources.base import ServiceResource
from boto3.session import Session
from botocore.client import BaseClient
from botocore.config import Config
from botocore.paginate import Paginator
from pandas import DataFrame

from docling.backend.docling_parse_v4_backend import DoclingParseV4DocumentBackend
from docling.backend.pdf_backend import PdfDocumentBackend
from docling.datamodel.base_models import ConversionStatus, InputFormat
from docling.datamodel.document import ConversionResult
from docling.datamodel.pipeline_options import PdfPipelineOptions
from docling.document_converter import DocumentConverter, PdfFormatOption
from docling.utils.utils import create_hash
from docling_core.types.doc.base import ImageRefMode
from docling_core.types.doc.document import (
    DocItem,
    DoclingDocument,
    PageItem,
    PictureClassificationData,
    PictureItem,
)
from docling_core.types.doc.labels import DocItemLabel

from docling_jobkit.datamodel.s3_inputs import S3Coordinates

logging.basicConfig(level=logging.INFO)

# Set the maximum file size of parquet to 500MB
MAX_PARQUET_FILE_SIZE = 500 * 1024 * 1024

classifier_labels = [
    "bar_chart",
    "bar_code",
    "chemistry_markush_structure",
    "chemistry_molecular_structure",
    "flow_chart",
    "icon",
    "line_chart",
    "logo",
    "map",
    "other",
    "pie_chart",
    "qr_code",
    "remote_sensing",
    "screenshot",
    "signature",
    "stamp",
]


def get_s3_connection(coords: S3Coordinates):
    session = Session()

    config = Config(
        connect_timeout=30, retries={"max_attempts": 1}, signature_version="s3v4"
    )
    scheme = "https" if coords.verify_ssl else "http"
    path = "/"
    endpoint = urlunsplit((scheme, coords.endpoint, path, "", ""))

    client: BaseClient = session.client(
        "s3",
        endpoint_url=endpoint,
        verify=coords.verify_ssl,
        aws_access_key_id=coords.access_key.get_secret_value(),
        aws_secret_access_key=coords.secret_key.get_secret_value(),
        config=config,
    )

    resource: ServiceResource = session.resource(
        "s3",
        endpoint_url=endpoint,
        verify=coords.verify_ssl,
        aws_access_key_id=coords.access_key.get_secret_value(),
        aws_secret_access_key=coords.secret_key.get_secret_value(),
        config=config,
    )

    return client, resource


def count_s3_objects(paginator: Paginator, bucket_name: str, prefix: str):
    response_iterator = paginator.paginate(Bucket=bucket_name, Prefix=prefix)
    count_obj = 0
    for page in response_iterator:
        if page.get("Contents"):
            count_obj += sum(1 for _ in page["Contents"])

    return count_obj


def get_keys_s3_objects_as_set(
    s3_resource: ServiceResource, bucket_name: str, prefix: str
):
    bucket = s3_resource.Bucket(bucket_name)
    folder_objects = list(bucket.objects.filter(Prefix=prefix))
    files_on_s3 = set()
    for file in folder_objects:
        files_on_s3.add(file.key)
    return files_on_s3


def strip_prefix_postfix(source_set: set, prefix: str = "", extension: str = ""):
    output = set()
    for key in source_set:
        output.add(key.replace(extension, "").replace(prefix, ""))
    return output


def generate_batch_keys(
    source_keys: list,
    batch_size: int = 10,
):
    batched_keys = []
    counter = 0
    sub_array = []
    array_lenght = len(source_keys)
    for idx, key in enumerate(source_keys):
        sub_array.append(key)
        counter += 1
        if counter == batch_size or (idx + 1) == array_lenght:
            batched_keys.append(sub_array)
            sub_array = []
            counter = 0

    return batched_keys


def generate_presign_url(
    s3_client: BaseClient,
    source_key: str,
    s3_source_bucket: str,
    expiration_time: int = 3600,
) -> str | None:
    try:
        return s3_client.generate_presigned_url(
            ClientMethod="get_object",
            Params={"Bucket": s3_source_bucket, "Key": source_key},
            ExpiresIn=expiration_time,
        )
    except Exception as e:
        logging.error("Generation of presigned url failed", exc_info=e)
        return None


def get_source_files(
    s3_source_client: BaseClient,
    s3_source_resource: ServiceResource,
    s3_coords: S3Coordinates,
):
    source_paginator = s3_source_client.get_paginator("list_objects_v2")

    key_prefix = (
        s3_coords.key_prefix
        if s3_coords.key_prefix.endswith("/")
        else s3_coords.key_prefix + "/"
    )
    # Check that source is not empty
    source_count = count_s3_objects(source_paginator, s3_coords.bucket, key_prefix)
    if source_count == 0:
        logging.error("No documents to process in the source s3 coordinates.")
    return get_keys_s3_objects_as_set(s3_source_resource, s3_coords.bucket, key_prefix)


def check_target_has_source_converted(
    coords: S3Coordinates,
    source_objects_list: list,
    s3_source_prefix: str,
):
    s3_target_client, s3_target_resource = get_s3_connection(coords)
    target_paginator = s3_target_client.get_paginator("list_objects_v2")

    converted_prefix = (
        coords.key_prefix + "json/"
        if coords.key_prefix.endswith("/")
        else coords.key_prefix + "/json/"
    )

    target_count = count_s3_objects(target_paginator, coords.bucket, converted_prefix)
    logging.debug("Target contains json objects: {}".format(target_count))
    if target_count != 0:
        logging.debug("Target contains objects, checking content...")

        # Collect target keys for iterative conversion
        existing_target_objects = get_keys_s3_objects_as_set(
            s3_target_resource, coords.bucket, converted_prefix
        )

        # Filter-out objects that are already processed
        target_short_key_list = strip_prefix_postfix(
            existing_target_objects, prefix=converted_prefix, extension=".json"
        )
        filtered_source_keys = []
        logging.debug("List of source keys:")
        for key in source_objects_list:
            logging.debug("Object key: {}".format(key))
            clean_key = key.replace(".pdf", "").replace(s3_source_prefix + "/", "")
            if clean_key not in target_short_key_list:
                filtered_source_keys.append(key)

        logging.debug("Total keys: {}".format(len(source_objects_list)))
        logging.debug("Filtered keys to process: {}".format(len(filtered_source_keys)))
    else:
        filtered_source_keys = source_objects_list

    return filtered_source_keys


def put_object(
    client: BaseClient,
    bucket: str,
    object_key: str,
    file: str,
    content_type: str | None = None,
) -> bool:
    """Upload a object to an S3 bucket

    :param file: Object to upload
    :param bucket: Bucket to upload to
    :param object_key: S3 key to upload to
    :return: True if object was uploaded, else False
    """

    kwargs = {}

    if content_type is not None:
        kwargs["ContentType"] = content_type

    try:
        client.put_object(Body=file, Bucket=bucket, Key=object_key, **kwargs)
    except Exception as e:
        logging.error("Put s3 object failed", exc_info=e)
        return False
    return True


def upload_file(
    client: BaseClient,
    bucket: str,
    object_key: str,
    file_name: str | Path,
    content_type: str | None = None,
):
    """Upload a file to an S3 bucket

    :param file_name: File to upload
    :param bucket: Bucket to upload to
    :param object_key: S3 key to upload to
    :param Optional[content_type]: Content type of file
    :return: True if file was uploaded, else False
    """

    kwargs = {}

    if content_type is not None:
        kwargs["ContentType"] = content_type

    try:
        client.upload_file(file_name, bucket, object_key, ExtraArgs={**kwargs})
    except Exception as e:
        logging.error("Upload file to s3 failed", exc_info=e)
        return False
    return True


class DoclingConvert:
    def __init__(
        self,
        source_s3_coords: S3Coordinates,
        target_s3_coords: S3Coordinates,
        pipeline_options: PdfPipelineOptions,
        allowed_formats: list[str] | None = None,
        to_formats: list[str] | None = None,
        backend: type[PdfDocumentBackend] | None = None,
        export_parquet_file: bool = True,
    ):
        self.source_coords = source_s3_coords
        self.source_s3_client, _ = get_s3_connection(source_s3_coords)

        self.target_coords = target_s3_coords
        self.target_s3_client, _ = get_s3_connection(target_s3_coords)

        self.converter = DocumentConverter(
            format_options={
                InputFormat.PDF: PdfFormatOption(
                    pipeline_options=pipeline_options,
                    backend=backend if backend else DoclingParseV4DocumentBackend,
                )
            }
        )
        if not allowed_formats:
            self.allowed_formats = [ext.value for ext in InputFormat]
        else:
            self.allowed_formats = [
                ext.value for ext in InputFormat if ext.value in allowed_formats
            ]

        self.to_formats = to_formats

        self.export_page_images = pipeline_options.generate_page_images
        self.export_images = pipeline_options.generate_picture_images
        self.export_parquet_file = export_parquet_file

        self.max_file_size = 1073741824  # TODO: be set from ENV

    def convert_documents(self, object_keys):
        pd_d = DataFrame()  # DataFrame to append parquet info
        try:
            for i, key in enumerate(object_keys):
                url = generate_presign_url(
                    self.source_s3_client,
                    key,
                    self.source_coords.bucket,
                    expiration_time=36000,
                )
                if not url:
                    continue
                parsed = urlparse(url)
                root, ext = os.path.splitext(parsed.path)
                # This will skip http links that don't have file extension as part of url, arXiv have plenty of docs like this
                if ext[1:] not in self.allowed_formats:
                    continue
                with tempfile.TemporaryDirectory() as tmpdirname:
                    temp_dir = Path(tmpdirname)
                    try:  # download file to be able to upload it later to s3
                        file_name = temp_dir / os.path.basename(parsed.path)
                        r = httpx.get(url, timeout=30)
                        content_length = r.headers.get("Content-Length", None)

                        if (
                            not content_length
                            or int(content_length) < self.max_file_size
                        ):
                            with open(file_name, "wb") as writer:
                                writer.write(r.content)
                    except Exception as exc:
                        logging.error(
                            "An error occurred downloading file.", exc_info=exc
                        )
                        yield f"{parsed.path} - FAILURE"
                        continue
                    try:
                        conv_res: ConversionResult = self.converter.convert(file_name)
                    except Exception as e:
                        logging.error(
                            "An error occurred while converting document.", exc_info=e
                        )
                        yield f"{parsed.path} - FAILURE"
                        continue
                    if conv_res.status == ConversionStatus.SUCCESS:
                        s3_target_prefix = self.target_coords.key_prefix
                        doc_hash = conv_res.input.document_hash
                        name_without_ext = os.path.splitext(file_name.name)[0]
                        logging.debug(f"Converted {doc_hash} now saving results")

                        if os.path.exists(conv_res.input.file):
                            self.upload_file_to_s3(
                                file=conv_res.input.file,
                                target_key=f"{s3_target_prefix}/pdf/{name_without_ext}.pdf",
                                content_type="application/pdf",
                            )

                        if self.export_page_images:
                            # Export pages images:
                            self.upload_page_images(
                                conv_res.document.pages,
                                s3_target_prefix,
                                conv_res.input.document_hash,
                            )

                        if self.export_images:
                            # Export pictures
                            self.upload_pictures(
                                conv_res.document,
                                s3_target_prefix,
                                conv_res.input.document_hash,
                            )

                        if self.to_formats is None or (
                            self.to_formats and "json" in self.to_formats
                        ):
                            # Export Docling document format to JSON:
                            target_key = (
                                f"{s3_target_prefix}/json/{name_without_ext}.json"
                            )
                            temp_json_file = temp_dir / f"{name_without_ext}.json"

                            conv_res.document.save_as_json(
                                filename=Path(temp_json_file.name),
                                image_mode=ImageRefMode.REFERENCED,
                            )
                            self.upload_file_to_s3(
                                file=temp_json_file.name,
                                target_key=target_key,
                                content_type="application/json",
                            )
                        if self.to_formats is None or (
                            self.to_formats and "doctags" in self.to_formats
                        ):
                            # Export Docling document format to doctags:
                            target_key = f"{s3_target_prefix}/doctags/{name_without_ext}.doctags.txt"

                            data = conv_res.document.export_to_document_tokens()
                            self.upload_object_to_s3(
                                file=data,
                                target_key=target_key,
                                content_type="text/plain",
                            )
                        if self.to_formats is None or (
                            self.to_formats and "md" in self.to_formats
                        ):
                            # Export Docling document format to markdown:
                            target_key = f"{s3_target_prefix}/md/{name_without_ext}.md"

                            data = conv_res.document.export_to_markdown()
                            self.upload_object_to_s3(
                                file=data,
                                target_key=target_key,
                                content_type="text/markdown",
                            )
                        if self.to_formats is None or (
                            self.to_formats and "html" in self.to_formats
                        ):
                            # Export Docling document format to html:
                            target_key = (
                                f"{s3_target_prefix}/html/{name_without_ext}.html"
                            )
                            temp_html_file = temp_dir / f"{name_without_ext}.html"

                            conv_res.document.save_as_html(temp_html_file)
                            self.upload_file_to_s3(
                                file=temp_html_file,
                                target_key=target_key,
                                content_type="text/html",
                            )

                        if self.to_formats is None or (
                            self.to_formats and "text" in self.to_formats
                        ):
                            # Export Docling document format to text:
                            target_key = (
                                f"{s3_target_prefix}/txt/{name_without_ext}.txt"
                            )

                            data = conv_res.document.export_to_text()
                            self.upload_object_to_s3(
                                file=data,
                                target_key=target_key,
                                content_type="text/plain",
                            )
                        if self.export_parquet_file:
                            logging.info("saving document info in dataframe...")
                            # Save Docling parquet info into DataFrame:
                            pd_d = self.document_to_dataframe(
                                conv_res=conv_res,
                                pd_dataframe=pd_d,
                                filename=name_without_ext,
                            )

                        yield f"{doc_hash} - SUCCESS"

                    elif conv_res.status == ConversionStatus.PARTIAL_SUCCESS:
                        yield f"{conv_res.input.file} - PARTIAL_SUCCESS"
                    else:
                        yield f"{conv_res.input.file} - FAILURE"

        finally:
            if self.export_parquet_file and not pd_d.empty:
                self.upload_parquet_file(pd_d, self.target_coords.key_prefix)

    def upload_object_to_s3(self, file, target_key, content_type):
        success = put_object(
            client=self.target_s3_client,
            bucket=self.target_coords.bucket,
            object_key=target_key,
            file=file,
            content_type=content_type,
        )
        if not success:
            logging.error(
                f"{file} - UPLOAD-FAIL: an error occour uploading object type {content_type} to s3"
            )
        return success

    def upload_file_to_s3(self, file, target_key, content_type):
        success = upload_file(
            client=self.target_s3_client,
            bucket=self.target_coords.bucket,
            object_key=target_key,
            file_name=file,
            content_type=content_type,
        )
        if not success:
            logging.error(
                f"{file} - UPLOAD-FAIL: an error occour uploading file type {content_type} to s3"
            )
        return success

    def upload_page_images(
        self,
        pages: dict[int, PageItem],
        s3_target_prefix: str,
        doc_hash: str,
    ):
        for page_no, page in pages.items():
            try:
                if page.image and page.image.pil_image:
                    page_hash = create_hash(f"{doc_hash}_page_no_{page_no}")
                    page_dpi = page.image.dpi
                    page_path_suffix = f"/pages/{page_hash}_{page_dpi}.png"
                    byteIO = BytesIO()
                    page.image.pil_image.save(byteIO, format="PNG")
                    self.upload_object_to_s3(
                        file=byteIO.getvalue(),
                        target_key=f"{s3_target_prefix}" + page_path_suffix,
                        content_type="application/png",
                    )
                    page.image.uri = Path(".." + page_path_suffix)

            except Exception as exc:
                logging.error(
                    "Upload image of page with hash %r raised error: %r",
                    page_hash,
                    exc,
                )

    def upload_pictures(
        self,
        document: DoclingDocument,
        s3_target_prefix: str,
        doc_hash: str,
    ):
        picture_number = 0
        for element, _level in document.iterate_items():
            if isinstance(element, PictureItem):
                if element.image and element.image.pil_image:
                    try:
                        element_hash = create_hash(f"{doc_hash}_img_{picture_number}")
                        element_dpi = element.image.dpi
                        element_path_suffix = (
                            f"/images/{element_hash}_{element_dpi}.png"
                        )
                        byteIO = BytesIO()
                        element.image.pil_image.save(byteIO, format="PNG")
                        self.upload_object_to_s3(
                            file=byteIO.getvalue(),
                            target_key=f"{s3_target_prefix}" + element_path_suffix,
                            content_type="application/png",
                        )
                        element.image.uri = Path(".." + element_path_suffix)

                    except Exception as exc:
                        logging.error(
                            "Upload picture with hash %r raised error: %r",
                            element_hash,
                            exc,
                        )
                    picture_number += 1

    def document_to_dataframe(
        self, conv_res: ConversionResult, pd_dataframe: DataFrame, filename: str
    ) -> DataFrame:
        result_table: list[dict[str, Any]] = []

        page_images = []
        for page_no, page in conv_res.document.pages.items():
            if page.image is not None and page.image.pil_image is not None:
                page_images.append(page.image.pil_image.tobytes())

        # Count the number of picture of each type
        num_formulas = 0
        num_codes = 0
        picture_classes = dict.fromkeys(classifier_labels, 0)
        for element, _level in conv_res.document.iterate_items():
            if isinstance(element, PictureItem):
                element.image = None  # reset images
                classification = next(
                    (
                        annot
                        for annot in element.annotations
                        if isinstance(annot, PictureClassificationData)
                    ),
                    None,
                )
                if classification is None or len(classification.predicted_classes) == 0:
                    continue

                predicted_class = classification.predicted_classes[0].class_name
                if predicted_class in picture_classes:
                    picture_classes[predicted_class] += 1

            elif isinstance(element, DocItem):
                if element.label == DocItemLabel.FORMULA:
                    num_formulas += 1
                elif element.label == DocItemLabel.CODE:
                    num_codes += 1

        num_pages = len(conv_res.document.pages)
        num_tables = len(conv_res.document.tables)
        num_elements = len(conv_res.document.texts)
        num_pictures = len(conv_res.document.pictures)

        # All features
        features = [
            num_pages,
            num_elements,
            num_tables,
            num_pictures,
            num_formulas,
            num_codes,
            *picture_classes.values(),
        ]

        doc_hash = (
            conv_res.document.origin.binary_hash
            if conv_res.document.origin
            else "unknown_hash"
        )
        doc_json = json.dumps(conv_res.document.export_to_dict())

        pdf_byte_array: bytearray | None = None
        if os.path.exists(conv_res.input.file):
            with open(conv_res.input.file, "rb") as file:
                pdf_byte_array = bytearray(file.read())

        result_table.append(
            {
                "filename": filename,
                "pdf": pdf_byte_array,
                "doc_hash": doc_hash,
                "document": doc_json,
                "page_images": page_images,
                "features": features,
                "doctags": str.encode(conv_res.document.export_to_document_tokens()),
            }
        )

        pd_df = pd.json_normalize(result_table)
        pd_df = pd_dataframe._append(pd_df)

        return pd_df

    def upload_parquet_file(self, pd_dataframe: DataFrame, s3_target_prefix: str):
        # Variables to track the file writing process
        file_index = 0
        current_file_size = 0
        current_df = pd.DataFrame()
        # Manifest dictionary
        manifest = {}
        # Current time
        timestamp = datetime.now().strftime("%Y-%m-%dT%H:%M:%SZ")

        while len(pd_dataframe) > 0:
            # Get a chunk of the DataFrame that fits within the file size limit
            chunk_size = min(
                len(pd_dataframe), MAX_PARQUET_FILE_SIZE // (current_file_size + 1)
            )

            # If the chunk size is 0, it means the current file size has exceeded the limit
            if chunk_size == 0:
                with tempfile.NamedTemporaryFile(
                    suffix=f".parquet_{file_index}"
                ) as temp_file:
                    pd_dataframe.to_parquet(temp_file.name)
                    current_file_size += os.path.getsize(temp_file.name)
                    file_index += 1

                    parquet_file_name = f"{timestamp}_{file_index}.parquet"
                    target_key = f"{s3_target_prefix}/parquet/{parquet_file_name}"
                    self.upload_file_to_s3(
                        file=temp_file.name,
                        target_key=target_key,
                        content_type="application/vnd.apache.parquet",
                    )

                    manifest[f"{parquet_file_name}"] = {
                        "filename": pd_dataframe["filename"].tolist(),
                        "doc_hash": pd_dataframe["doc_hash"].tolist(),
                        "row_number": 3,
                        "timestamp": timestamp,
                    }

                pd_dataframe = pd.DataFrame()
            else:
                # Get the current chunk of the DataFrame
                current_df = pd_dataframe.iloc[:chunk_size]
                pd_dataframe = pd_dataframe.iloc[chunk_size:]

                with tempfile.NamedTemporaryFile(
                    suffix=f".parquet_{file_index}"
                ) as temp_file:
                    current_df.to_parquet(temp_file.name)
                    current_file_size += os.path.getsize(temp_file.name)
                    file_index += 1

                    parquet_file_name = f"{timestamp}_{file_index}.parquet"
                    target_key = f"{s3_target_prefix}/parquet/{parquet_file_name}"
                    self.upload_file_to_s3(
                        file=temp_file.name,
                        target_key=target_key,
                        content_type="application/vnd.apache.parquet",
                    )

                    manifest[f"{parquet_file_name}"] = {
                        "filenames": current_df["filename"].tolist(),
                        "doc_hashes": current_df["doc_hash"].tolist(),
                        "row_number": 3,
                        "timestamp": timestamp,
                    }

        logging.info(f"Total parquet files uploaded: {file_index}")

        # Export manifest file:
        with tempfile.NamedTemporaryFile(suffix=".json") as temp_file_json:
            with open(temp_file_json.name, "w") as file:
                json.dump(manifest, file, indent=4)
            self.upload_file_to_s3(
                file=temp_file_json.name,
                target_key=f"{s3_target_prefix}/manifest/{timestamp}.json",
                content_type="application/json",
            )



from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from docling_jobkit.orchestrators.base_orchestrator import BaseOrchestrator


class BaseNotifier(ABC):
    def __init__(self, orchestrator: "BaseOrchestrator"):
        self.orchestrator = orchestrator

    @abstractmethod
    async def add_task(self, task_id: str):
        pass

    @abstractmethod
    async def remove_task(self, task_id: str):
        pass

    @abstractmethod
    async def notify_task_subscribers(self, task_id: str):
        pass

    @abstractmethod
    async def notify_queue_positions(self):
        pass



# ruff: noqa: E402

from typing import List, NamedTuple, cast

from kfp import dsl
from kfp.dsl import Dataset, Input, Output


@dsl.component(
    packages_to_install=[
        "docling==2.28.0",
        "git+https://github.com/docling-project/docling-jobkit@2c27c71b75da98f04fccc7abc7ddc3a9a3afb0cd",
    ],
    base_image="quay.io/docling-project/docling-serve:jobkit-base-0.0.19",  # base docling-serve image with fixed permissions
)
def convert_payload(
    options: dict,
    source: dict,
    target: dict,
    batch_index: int,
    # source_keys: List[str],
    dataset: Input[Dataset],
) -> list:
    import json
    import logging
    import os
    from typing import Optional

    from docling.backend.docling_parse_backend import DoclingParseDocumentBackend
    from docling.backend.docling_parse_v2_backend import DoclingParseV2DocumentBackend
    from docling.backend.docling_parse_v4_backend import DoclingParseV4DocumentBackend
    from docling.backend.pdf_backend import PdfDocumentBackend
    from docling.backend.pypdfium2_backend import PyPdfiumDocumentBackend
    from docling.datamodel.pipeline_options import (
        OcrOptions,
        PdfBackend,
        PdfPipelineOptions,
        TableFormerMode,
    )
    from docling.models.factories import get_ocr_factory

    from docling_jobkit.connectors.s3_helper import DoclingConvert
    from docling_jobkit.datamodel.convert import ConvertDocumentsOptions
    from docling_jobkit.datamodel.s3_inputs import S3Coordinates

    logging.basicConfig(level=logging.INFO)

    # set expected path to pre-loaded models
    os.environ["DOCLING_ARTIFACTS_PATH"] = "/opt/app-root/src/.cache/docling/models"
    # easyocr_path = Path("/opt/app-root/src/.cache/docling/models/EasyOcr")
    # os.environ["MODULE_PATH"] = str(easyocr_path)
    # os.environ["EASYOCR_MODULE_PATH"] = str(easyocr_path)

    # validate inputs
    source_s3_coords = S3Coordinates.model_validate(source)
    target_s3_coords = S3Coordinates.model_validate(target)

    convert_options = ConvertDocumentsOptions.model_validate(options)

    backend: Optional[type[PdfDocumentBackend]] = None
    if convert_options.pdf_backend:
        if convert_options.pdf_backend == PdfBackend.DLPARSE_V1:
            backend = DoclingParseDocumentBackend
        elif convert_options.pdf_backend == PdfBackend.DLPARSE_V2:
            backend = DoclingParseV2DocumentBackend
        elif convert_options.pdf_backend == PdfBackend.DLPARSE_V4:
            backend = DoclingParseV4DocumentBackend
        elif convert_options.pdf_backend == PdfBackend.PYPDFIUM2:
            backend = PyPdfiumDocumentBackend
        else:
            raise RuntimeError(
                f"Unexpected PDF backend type {convert_options.pdf_backend}"
            )

    pipeline_options = PdfPipelineOptions()
    pipeline_options.do_ocr = convert_options.do_ocr
    ocr_factory = get_ocr_factory()

    pipeline_options.ocr_options = cast(
        OcrOptions, ocr_factory.create_options(kind=convert_options.ocr_engine)
    )

    pipeline_options.do_table_structure = convert_options.do_table_structure
    pipeline_options.table_structure_options.mode = TableFormerMode(
        convert_options.table_mode
    )
    pipeline_options.generate_page_images = convert_options.include_images
    pipeline_options.do_code_enrichment = convert_options.do_code_enrichment
    pipeline_options.do_formula_enrichment = convert_options.do_formula_enrichment
    pipeline_options.do_picture_classification = (
        convert_options.do_picture_classification
    )
    pipeline_options.do_picture_description = convert_options.do_picture_description
    pipeline_options.generate_picture_images = convert_options.include_images

    # pipeline_options.accelerator_options = AcceleratorOptions(
    #     num_threads=2, device=AcceleratorDevice.CUDA
    # )

    converter = DoclingConvert(
        source_s3_coords=source_s3_coords,
        target_s3_coords=target_s3_coords,
        pipeline_options=pipeline_options,
        allowed_formats=[str(v) for v in convert_options.from_formats],
        to_formats=[str(v) for v in convert_options.to_formats],
        backend=backend,
    )

    with open(dataset.path) as f:
        batches = json.load(f)
    source_keys = batches[batch_index]

    results = []
    for item in converter.convert_documents(source_keys):
        results.append(item)
        logging.info("Convertion result: {}".format(item))

    return results


@dsl.component(
    packages_to_install=[
        "pydantic",
        "boto3~=1.35.36",
        "git+https://github.com/docling-project/docling-jobkit@2c27c71b75da98f04fccc7abc7ddc3a9a3afb0cd",
    ],
    base_image="python:3.11",
)
def compute_batches(
    source: dict,
    target: dict,
    dataset: Output[Dataset],
    batch_size: int = 10,
) -> NamedTuple("outputs", [("batch_indices", List[int])]):  # type: ignore[valid-type]
    import json
    from typing import NamedTuple

    from docling_jobkit.connectors.s3_helper import (
        check_target_has_source_converted,
        generate_batch_keys,
        get_s3_connection,
        get_source_files,
    )
    from docling_jobkit.datamodel.s3_inputs import S3Coordinates

    # validate inputs
    s3_coords_source = S3Coordinates.model_validate(source)
    s3_target_coords = S3Coordinates.model_validate(target)

    s3_source_client, s3_source_resource = get_s3_connection(s3_coords_source)
    source_objects_list = get_source_files(
        s3_source_client, s3_source_resource, s3_coords_source
    )
    filtered_source_keys = check_target_has_source_converted(
        s3_target_coords, source_objects_list, s3_coords_source.key_prefix
    )
    batch_keys = generate_batch_keys(
        filtered_source_keys,
        batch_size=batch_size,
    )

    with open(dataset.path, "w") as out_batches:
        json.dump(batch_keys, out_batches)

    batch_indices = list(range(len(batch_keys)))
    outputs = NamedTuple("outputs", [("batch_indices", List[int])])
    return outputs(batch_indices)


@dsl.pipeline
def inputs_s3in_s3out(
    convertion_options: dict = {
        "from_formats": [
            "docx",
            "pptx",
            "html",
            "image",
            "pdf",
            "asciidoc",
            "md",
            "xlsx",
            "xml_uspto",
            "xml_jats",
            "json_docling",
        ],
        "to_formats": ["md", "json", "html", "text", "doctags"],
        "image_export_mode": "placeholder",
        "do_ocr": True,
        "force_ocr": False,
        "ocr_engine": "easyocr",
        "ocr_lang": [],
        "pdf_backend": "dlparse_v2",
        "table_mode": "accurate",
        "abort_on_error": False,
        "return_as_file": False,
        "do_table_structure": True,
        "do_code_enrichment": False,
        "do_formula_enrichment": False,
        "do_picture_classification": False,
        "do_picture_description": False,
        "generate_picture_images": False,
        "include_images": True,
        "images_scale": 2,
    },
    source: dict = {
        "endpoint": "s3.eu-de.cloud-object-storage.appdomain.cloud",
        "access_key": "123454321",
        "secret_key": "secretsecret",
        "bucket": "source-bucket",
        "key_prefix": "my-docs",
        "verify_ssl": True,
    },
    target: dict = {
        "endpoint": "s3.eu-de.cloud-object-storage.appdomain.cloud",
        "access_key": "123454321",
        "secret_key": "secretsecret",
        "bucket": "target-bucket",
        "key_prefix": "my-docs",
        "verify_ssl": True,
    },
    batch_size: int = 20,
):
    import logging

    logging.basicConfig(level=logging.INFO)

    batches = compute_batches(source=source, target=target, batch_size=batch_size)
    # disable caching on batches as cached pre-signed urls might have already expired
    batches.set_caching_options(False)

    results = []
    with dsl.ParallelFor(batches.outputs["batch_indices"], parallelism=20) as subbatch:
        converter = convert_payload(
            options=convertion_options,
            source=source,
            target=target,
            dataset=batches.outputs["dataset"],
            batch_index=subbatch,
        )
        converter.set_memory_request("1G")
        converter.set_memory_limit("7G")
        converter.set_cpu_request("200m")
        converter.set_cpu_limit("1")

        # For enabling document conversion using GPU
        # currently unable to properly pass input parameters into pipeline, therefore node selector and tolerations are hardcoded

        # converter.set_accelerator_type("nvidia.com/gpu")
        # converter.set_accelerator_limit("1")

        # kubernetes.add_node_selector(
        #     task=converter,
        #     label_key="nvidia.com/gpu.product",
        #     label_value="NVIDIA-A10",
        # )

        # kubernetes.add_toleration(
        #     task=converter,
        #     key="key1",
        #     operator="Equal",
        #     value="mcad",
        #     effect="NoSchedule",
        # )

        results.append(converter.output)


### Compile pipeline into a yaml
from kfp import compiler

compiler.Compiler().compile(inputs_s3in_s3out, "docling-s3in-s3out.yaml")


### Start pipeline run programatically
# import kfp
# import os

# # TIP: you may need to authenticate with the KFP instance
# kfp_client = kfp.Client(
#     host = os.environ["KFP_FULL_URL"],
#     existing_token = os.environ["OPENSHIFT_TOKEN"],
#     verify_ssl = False,
# )

# kfp_client.create_run_from_pipeline_func(
#     inputs_s3in_s3out,
#     arguments=dict(
#         convertion_options = {
#             "from_formats": [
#                 "pdf",
#             ],
#             "to_formats": ["md", "json", "html", "text", "doctags"],
#             "image_export_mode": "placeholder",
#             "do_ocr": True,
#             "force_ocr": True,
#             "ocr_engine": "easyocr",
#         },
#         source = {
#             "endpoint": os.environ["S3_SOURCE_ENDPOINT"],
#             "access_key": os.environ["S3_SOURCE_ACCESS_KEY"],
#             "secret_key": os.environ["S3_SOURCE_SECRET_KEY"],
#             "bucket": os.environ["S3_SOURCE_BUCKET"],
#             "key_prefix": os.environ["S3_SOURCE_PREFIX"],
#             "verify_ssl": True,
#         },
#         target = {
#             "endpoint": os.environ["S3_TARGET_ENDPOINT"],
#             "access_key": os.environ["S3_TARGET_ACCESS_KEY"],
#             "secret_key": os.environ["S3_TARGET_SECRET_KEY"],
#             "bucket": os.environ["S3_TARGET_BUCKET"],
#             "key_prefix": os.environ["S3_TARGET_PREFIX"],
#             "verify_ssl": True,
#         },
#         batch_size = 100,
#     )
# )



import enum

from pydantic import BaseModel


class TaskStatus(str, enum.Enum):
    SUCCESS = "success"
    PENDING = "pending"
    STARTED = "started"
    FAILURE = "failure"


class TaskProcessingMeta(BaseModel):
    num_docs: int
    num_processed: int = 0
    num_succeeded: int = 0
    num_failed: int = 0
